-- Operations with fractional numbers
-- You may free use and modify this file 
-- without deleting this comment group
-- Copyright (c) 1998 Dimitriy Anisimkov
-- e-mail anisimkov@yahoo.com
-- phone +7 (381)-2-217774
package body Rational is

   function Trunc(R : TRational) return TRational is
   begin
      return (Numer=>R.Numer/R.Denom,Denom=>Onecomp);
   end;

   function Euclide(A,B : Comp;Bl : Boolean) return TRational is
   begin
      if (B=Onecomp) then
         if not Bl then
            return (Numer => A-Onecomp,Denom=>B);
         else
            return Unlimited;
         end if;
      end if;
      return Sumrc(Inverse(Euclide(B,A rem B,not Bl)),A/B);
   end;

   procedure Chknd(R : TRational) is
      Negative_Denominator : exception;
   begin
      if R.Denom<Nilcomp then
         raise Negative_Denominator;
      end if;
   end;

   function LowestFactor(C1,C2 : Comp) return Comp is
   begin
      return C1/GreatestDivisor(C1,C2)*C2;
   end;

--   function GreatestDivisor(C1,C2 : Comp) return Comp is
--   begin
--      if C2=Nilcomp then
--         return abs(C1);
--      else
--         return GreatestDivisor(C2,C1 rem C2);
--      end if;
--   end;

   function GreatestDivisor(C1,C2 : Comp) return Comp is
      B1 : Comp := C1;
      B2 : Comp := C2;
   begin
      while B2/=Nilcomp loop
         B1 := B1 rem B2;
         if B1=Nilcomp then
            return abs(B2);
         end if;
         B2 := B2 rem B1;
      end loop;
      return abs(B1);
   end;

   function MinMembBetw(R1,R2 : TRational) return TRational is
      -- r1>=NilComp,r2>=NilComp,r1<{=?}r2
      C,C2 : Comp;
   begin
      C := R1.Numer/R1.Denom;
      C2 := R2.Numer/R2.Denom;
      if C<C2 then
         return (Numer=>C+OneComp,Denom=>OneComp);
      else
         return SumRC(Inverse(MinMembBetw(
                  Inverse(SumRC(R2,-C)),Inverse(SumRC(R1,-C))
                  )),C);
      end if;
   end;

   function Fract(R : TRational) return TRational is
   begin
      return (Numer=>R.Numer rem R.Denom,Denom=>R.Denom);
   end;

   procedure ReduceComps(C1,C2 : in out Comp) is
      Cdev : Comp;
   begin
      Cdev := Greatestdivisor(C1,C2);
      C1 := C1/Cdev;
      C2 := C2/Cdev;
   end;

   function Reduced(R : TRational) return TRational is
      Cdev : Comp;
   begin
      Cdev := Greatestdivisor(R.Numer,R.Denom);
      return (Numer=>R.Numer/Cdev,Denom=>R.Denom/Cdev);
   end;

   function "/"(Num,Denom : Comp) return TRational is
      Cdev : Comp;
   begin
      if Denom=nilComp then
        return (Numer=>Num,Denom=>Denom);
      end if;
      Cdev := GreatestDivisor(Num,Denom);
      if Denom>Nilcomp then
         return (Numer=>Num/Cdev,Denom=>Denom/Cdev);
      else
         return (Numer=>-Num/Cdev,Denom=>-Denom/Cdev);
      end if;
   end;

   function SumRC(R : TRational;C : Comp) return TRational is
   begin
      return (Numer=>C*R.Denom+R.Numer,Denom=>R.Denom);
   end;

   function "+"(R1,R2 : TRational) return TRational is
   begin
      return (R1.Numer*R2.Denom+R2.Numer*R1.Denom)/(R1.Denom*R2.Denom);
   end;

   function "-"(R1,R2 : TRational) return TRational is
   begin
      return (R1.Numer*R2.Denom-R2.Numer*R1.Denom)/(R1.Denom*R2.Denom);
   end;

   function "*"(R1,R2 : TRational) return TRational is
   begin
      return (R1.Numer*R2.Numer)/(R1.Denom*R2.Denom);
   end;

   function "/"(R1,R2 : TRational) return TRational is
   begin
      return R1*Inverse(R2);
   end;

   function MulRC(R : TRational;C : Comp) return TRational is
   begin
      return (R.Numer*C)/R.Denom;
   end;

   function Inverse(R : TRational) return TRational is
   begin
      if R.Numer<Nilcomp then
         return (Numer=>-R.Denom,Denom=>-R.Numer);
      else
         return (Numer=>R.Denom,Denom=>R.Numer);
      end if;
   end;

   function "-"(R : TRational) return TRational is
   begin
      return (Numer=>-R.Numer,Denom=>R.Denom);
   end;

   generic
      with function Compare(C1,C2 : Comp) return Boolean;
   function Compare(R1,R2 : TRational) return Boolean;

   function Compare(R1,R2 : TRational) return Boolean is
   begin
      return Compare(R1.Numer*R2.Denom,R2.Numer*R1.Denom);
   end;

   function Lower is new Compare("<");
   function More is new Compare(">");
   function Lowereq is new Compare("<=");
   function Moreeq is new Compare(">=");
   function Equal is new Compare("=");

   function "<"(R1,R2 : TRational) return Boolean renames Lower;
   function ">"(R1,R2 : TRational) return Boolean renames More;
   function "<="(R1,R2 : TRational) return Boolean renames Lowereq;
   function ">="(R1,R2 : TRational) return Boolean renames Moreeq;
   -- AONIX build 7.1.424 can't make function "=" directly 
   -- look to the specification
   function Equal4Aonix(R1,R2 : TRational) return Boolean renames Equal;

end Rational;