-- Unlimited number static implementation
-- You may use and modify this file free
-- without deleting this comment group
-- Copyright (c) 1999 Dimitriy Anisimkov
-- e-mail anisimkov@yahoo.com
with Ada.Unchecked_Conversion;
package body Unlimited.Number is
   use Interfaces;

   function P is new Ada.Unchecked_Conversion(UInt,Part);
   function U is new Ada.Unchecked_Conversion(Part,UInt);

   function Shift_Range(A : Unlimited_Static; Sh : Integer) return Unlimited_Static;
   function Sign_Part(Arr : Unlimited_Static) return Part;
   function Is_Null(U : Unlimited_Static) return Boolean;
   function Extend_First(Ui : Unlimited_Static;First : Integer) return Unlimited_Static;
   function Extend_Last(Ui : Unlimited_Static;Last : Integer) return Unlimited_Static;
   function Extend_Range(Ui : Unlimited_Static;First,Last : Integer) return Unlimited_Static;
   procedure Assign(Left : out Unlimited_Static; Right : in Unlimited_Static);
   -- pragma Inline(Shift_Range,Sign_Part,Is_Null,Extend_First,Extend_Last,Extend_Range,Assign);

   function Integer2unlimited(I : Int) return Unlimited_Static is
   begin
      return (0 => Part(I));
   end;

   function Sign_Part(Arr : Unlimited_Static) return Part is
   begin
      return -Boolean'Pos(Int(Arr(Arr'Last))<0);
   end;

   generic
      with function Oper(A1,A2 : LLI) return LLI;
   procedure Oper_Uni(Res : in out Unlimited_Static;AU : in Unlimited_Static);
   procedure Oper_Uni(Res : in out Unlimited_Static;AU : in Unlimited_Static) is
      Rli : LLI;
      Sign : Part;
   begin
      Set_High(Rli,0);
      for I in AU'range loop
         Rli := Oper(To_Lli(U(Res(I))),To_Lli(U(AU(I))))+To_Lli(High(Rli));
         Res(I) := Low(Rli);
      end loop;
      Sign := Sign_Part(AU);
      for I in AU'Last+1..Res'Last loop
         Rli := Oper(To_Lli(U(Res(I))),To_Lli(U(Sign)))+To_Lli(High(Rli));
         Res(I) := Low(Rli);
         exit when High(Rli)=0 and Sign=0;
      end loop;
   end;

   procedure Add is new Oper_Uni("+");
   procedure Sub is new Oper_Uni("-");

   procedure Reducelast(A : in Unlimited_Static;First : in Integer;Last : in out Integer) is
   begin
      while First<Last and then A(Last)=0 and then (A(Last-1)>=0) loop
         Last := Last-1;
      end loop;
      while First<Last and then A(Last)=-1 and then A(Last-1)<0 loop
         Last := Last-1;
      end loop;
   end;

   procedure Reducefirst(A : in Unlimited_Static;First : in out Integer;Last : in Integer) is
   begin
      while A(First)=0 and then First<Last loop
         First := First+1;
      end loop;
   end;

   function Reduce(A : in Unlimited_Static) return Unlimited_Static is
      First : Integer := A'First;
      Last : Integer := A'Last;
   begin
      Reducelast(A,First,Last);
      Reducefirst(A,First,Last);
      return A(First..Last);
   end;

   function Is_Null(U : Unlimited_Static) return Boolean is
      First : Integer := U'First;
   begin
      Reducefirst(U,First,U'Last);
      return U(First)=0;
   end;

   function Extend_First(Ui : Unlimited_Static;First : Integer) return Unlimited_Static is
   begin
      return (First..Ui'First-1 => 0)&Ui;
   end;

   function Extend_Last(Ui : Unlimited_Static;Last : Integer) return Unlimited_Static is
   begin
      return Ui&(Ui'Last+1..Last => Sign_Part(Ui));
   end;

   function Extend_Range(Ui : Unlimited_Static;First,Last : Integer) return Unlimited_Static is
   begin
      if Is_Null(Ui) then
         return (First..Last=>0);
      else
         return Extend_First(Extend_Last(Ui,Last),First);
      end if;
   end;

   procedure Assign(Left : out Unlimited_Static; Right : in Unlimited_Static) is
   begin
      Left := Extend_Range(Right,Left'First,Left'Last);
   end;

   function "*"(Left : in Unlimited_Static;Right : Int) return Unlimited_Static is
      Res : Unlimited_Static(Left'First..Left'Last+1);
      Rli : LLI := Int(Left(Left'Last))*Right;
   begin
      Res(Left'Last) := Low(Rli);
      Res(Left'Last+1) := High(Rli);
      for Idx in reverse Left'First..Left'Last-1 loop
         Rli := U(Left(Idx))*Right;
         Res(Idx) := Low(Rli);
         Add(Res, (Idx+1 => High(Rli)));
      end loop;
      return Reduce(Res);
   end;

   function Shift_Range(A : Unlimited_Static; Sh : Integer) return Unlimited_Static is
      Result : Unlimited_Static(A'First+Sh..A'Last+Sh) := A;
   begin
      return Result;
   end;

   function "*"(A : in Unlimited_Static;Ui : UInt) return Unlimited_Static is
   function S is new Ada.Unchecked_Conversion(UInt,Int);
   begin
      if S(Ui)>=0 then
         return A*S(Ui);
      else
         return A*S(Ui)+Shift_Range(A,1);
      end if;
   end;

   function Bits_In(N : in Part) return Natural is
      Res : Natural := 0;
      Sh : Int := Int(N);
   begin
      while Sh/=0 loop
         Sh := Sh/2;
         Res := Res+1;
      end loop;
      return Res;
   end;

   function "*"(Left,Right : in Unlimited_Static) return Unlimited_Static is
      Resfirst : constant Integer := Left'First+Right'First;
      Res : Unlimited_Static(Resfirst..Resfirst+Left'Length+Right'Length) := (others => 0);
      Left_Last : integer := Left'Last;
   begin
      ReduceLast(Left,Left'First,Left_Last);
      for I in Left'First..Left_Last-1 loop
         Add(Res,Shift_Range(Right,I)*U(Left(I)));
      end loop;
      Add(Res,Shift_Range(Right,Left_Last)*Int(Left(Left_Last)));
      return Reduce(Res);
   end;

   function To_String(Num : in Unlimited_Static; 
       Trans : in Conversion_Chars; 
       Power2Enabled : Boolean:=True) return String is
      Remain : Unlimited_Static(0..0);
      ShPwr : Integer := Boolean'Pos(Power2Enabled)*Num'First;
      Ai : Unlimited_Static := Shift_Range(Num,-ShPwr)&Sign_Part(Num);
      Result : String(1..Ai'Size/(Bits_In(Trans'Length)-1)) := (others => ' ');
      Idx : Integer := Result'Last;
      function PowerStr return String is
      begin
         if ShPwr/=0 then
            return "*2^"&Integer'Image(ShPwr*Int'Size);
         else
            return "";
         end if;
      end;
   begin
      loop
         Divide(Ai,Trans'Length,Ai,Remain);
         Result(Idx) := Trans(abs(Int(Remain(0))));
         exit when Is_Null(Ai);
         Idx := Idx-1;
      end loop;
      if Num(Num'Last)<0 then
         return "-"&Result(Idx..Result'Last)&Powerstr;
      else
         return Result(Idx..Result'Last)&Powerstr;
      end if;
   end;

   function To_String(Num : in Unlimited_Static;
         Base : Base_Of_Presentation :=10;Power2Enabled : Boolean:=True) return String is
   begin
      return To_String(Num,Hexdigs(0..Base-1),Power2Enabled);
   end;

   generic
      with function Compare(I1,I2 : Int) return Boolean;
      with function Compare(I1,I2 : Uint) return Boolean;
      with function Compare(I1,I2 : Integer) return Boolean;
   function Compare(Ui1,Ui2 : Unlimited_Static) return Boolean;

   function Compare(Ui1,Ui2 : Unlimited_Static) return Boolean is
      Last1 : Integer := Ui1'Last;
      Last2 : Integer := Ui2'Last;
      First1 : Integer := Ui1'First;
      First2 : Integer := Ui2'First;
   begin
      if (UI1(Last1)<0) xor (UI2(Last2)<0) then
         return Compare(Int(UI1(Last1)),Int(UI2(Last2)));
      end if;
      Reducefirst(Ui1,First1,Last1);
      ReduceFirst(Ui2,First2,Last2);
      if UI1(First1)=0 or UI2(First2)=0 then
         return Compare(U(UI1(First1)),U(UI2(First2)));
      end if;
      Reducelast(Ui1,First1,Last1);
      Reducelast(Ui2,First2,Last2);
      if Last1/=Last2 then
         if UI1(Last1)>=0 then
            return Compare(Last1,Last2);
         else
            return Compare(Last2,Last1);
         end if;
      end if;
      declare
         First : constant Integer := Integer'Max(First1,First2);
      begin
         for I in reverse First..Last1 loop
            if UI1(I)/=UI2(I) then
               return Compare(U(UI1(I)),U(UI2(I)));
            end if;
         end loop;
      end;
      return Compare(First2,First1);
   end Compare;

   function Lower is new Compare("<","<","<");
   function "<"(Left,Right : Unlimited_Static) return Boolean renames Lower;

   function More is new Compare(">",">",">");
   function ">"(Left,Right : Unlimited_Static) return Boolean renames More;

   function LowerEq is new Compare("<=","<=","<=");
   function "<="(Left,Right : Unlimited_Static) return Boolean renames LowerEq;

   function MoreEq is new Compare(">=",">=",">=");
   function ">="(Left,Right : Unlimited_Static) return Boolean renames MoreEq;

   function Equal is new Compare("=","=","=");
   function "="(Left,Right : Unlimited_Static) return Boolean renames Equal;

   function Shift(Ui : Unlimited_Static;Amount : Integer) return Unlimited_Static is
      Shi : constant Natural := Amount mod Int'Size;
      Shr : constant Integer := Amount/Int'Size-Boolean'Pos(Amount<0 and Shi/=0);
      Result : Unlimited_Static(Ui'First+Shr..Ui'Last+Shr+1);
      Buf : UInt;
      Next : UInt := 0;
      Mask : constant UInt := Shift_Left(1,Shi)-1;
   begin
      for I in Ui'range loop
         Buf := Rotate_Left(U(Ui(I)),Shi);
         Result(I+Shr) := P((Buf and not Mask) or Next);
         Next := Buf and Mask;
      end loop;
      Result(Result'Last) := P((U(Sign_Part(Ui)) and not Mask) or Next);
      return Reduce(Result);
   end Shift;

   function LowBits0(N : Part) return Natural is
      Result : Natural := Int'Size;
      Shi : Uint := U(N);
   begin
      while Shi/=0 loop
         Shi := Shift_Left(Shi,1);
         Result := Result-1;
      end loop;
      return Result;
   end;

   procedure Divide(Numerator,Denominator : in Unlimited_Static;
         Result,Remainder : out Unlimited_Static) is
      Denlast : Integer := Denominator'Last;
      SignN : Boolean := Numerator(Numerator'Last)<0;
      SignD : Boolean := Denominator(DenLast)<0;
      Shl : Integer;
   begin
      ReduceLast(Denominator,Denominator'First,Denlast);
      if Denominator=Nil then
         raise Division_By_Zero;
      end if;
      Shl := Int'Size-1-Bits_In(Denominator(Denlast));
      declare
         Denom : constant Unlimited_Static := abs(Shift(Denominator,Shl));
         Den : constant Int := Int(Denom(DenLast));
         Remain : Unlimited_Static := abs(Extend_First(Shift(Numerator,Shl),Denom'First))&0;
         Cidx : Integer := Remain'Last;
      begin
         Result := (others => 0);
         while Remain>=Denom loop
            ReduceLast(Remain,Remain'First,Cidx);
            if Cidx=Denlast then
               Cidx := Cidx+1;
            end if;
            declare
               Shi : constant Integer := Cidx-Denlast;
               D_Int : constant LLI := To_LLI(
                     Low=>Remain(Cidx-1),
                     High=>Remain(Cidx))/To_Lli(Den);
               --Parts(To_Lli(TwoInt(Remain(Cidx-1..Cidx))));
               Cma : Unlimited_Static(Shi-1..Shi);
            begin
               Cma(Shi-1) := Low(d_int);
               Cma(Shi) := High(d_int);
               Add(Result,Cma);
               Sub(Remain,Denom*Cma);
               while Remain(Remain'Last)<0 loop
                  Add(Remain,Shift_Range(Denom,Shi-1));
                  Add(Result,(Shi-1=>-1));
               end loop;
            end;
         end loop;
         if SignN then
           Assign(Remainder,Shift(-Remain,-Shl));
         else
           Assign(Remainder,Shift(Remain,-Shl));
         end if;
         if SignN xor SignD then
            Result := -Result;
         end if;
      end;
   end;

   procedure Divide(Numer : in Unlimited_Static; Denom : in Int;
         Result,Remainder : out Unlimited_Static) is
      Num : constant Unlimited_Static := Numer;
      Sign : constant Boolean := Num(Num'Last)<0;
      Rli,Rt : LLI;
      RemSign : Boolean;
      Error_Result_Range : exception;
   begin
--      Divide(Numer,(0=>part(Denom)),Result,Remainder);
--      return;
      if Numer'Last<0 then
         Result := (others => 0);
         Assign(Remainder,Num);
         return;
      end if;
      Result(Num'Last) := Int(Num(Num'Last))/Denom;
      Result(Num'Last+1..Result'Last) := (others => -Boolean'Pos(Int(Result(Num'Last))<0));
      Set_High(Rli, Int(Num(Num'Last)) rem Denom);
      for I in reverse 0..Num'Last-1 loop
         if I>=Num'First then
            Set_Low(Rli, Num(I));
         else
            Set_Low(Rli,0);
         end if;
         Rt := Rli/To_Lli(Denom);
         Result(I) := Low(Rt);
         Add(Result,(I+1=>high(Rt)));
         Set_high(Rli, Low(Rli rem To_Lli(Denom)));
      end loop;
      Remainder(0) := high(Rli);
      if Remainder'First<0 then
         Remainder(Remainder'First..-1) := Extend_First(Num(Num'First..-1),Remainder'First);
      end if;
      RemSign := Remainder<Nil;
      if (not Is_Null(Remainder)) and then (Sign xor RemSign) then
         if Denom<0 xor RemSign then
            Add(Remainder,(0=>Denom));
            Sub(Result,(0=>1));
         else
            Sub(Remainder,(0=>Denom));
            Add(Result,(0=>1));
         end if;
      end if;
   end;

   function "/"(Left,Right : in Unlimited_Static) return Unlimited_Static is
      Result : Unlimited_Static(0..Integer'Max(Left'Last-Right'First+1,0));
      Remain : Unlimited_Static(Integer'Min(Left'First,Right'First)..Integer'Min(Left'Last,Right'Last));
   begin
      Divide(Left,Right,Result,Remain);
      return Reduce(Result);
   end;

   function "rem"(Left,Right : in Unlimited_Static) return Unlimited_Static is
      Result : Unlimited_Static(0..Integer'Max(Left'Last-Right'First+1,0));
      Remain : Unlimited_Static(Integer'Min(Left'First,Right'First)..Integer'Min(Left'Last,Right'Last));
   begin
      Divide(Left,Right,Result,Remain);
      return Reduce(Remain);
   end;

   function "/"(Left : in Unlimited_Static;Right : Int) return Unlimited_Static is
      Result : Unlimited_Static(0..Integer'Max(Left'Last+1,0));
      Remain : Unlimited_Static(Integer'Min(Left'First,0)..0);
   begin
      Divide(Left,Right,Result,Remain);
      return Reduce(Result);
   end;

   function "rem"(Left : in Unlimited_Static;Right : Int) return Unlimited_Static is
      Result : Unlimited_Static(0..Integer'Max(Left'Last+1,0));
      Remain : Unlimited_Static(Integer'Min(Left'First,0)..0);
   begin
      Divide(Left,Right,Result,Remain);
      return Reduce(Remain);
   end;

   function "not"(Ui : in Unlimited_Static) return Unlimited_Static is
      Res : Unlimited_Static(Ui'range);
   begin
      for I in Ui'range loop
         Res(I) := P(not U(Ui(I)));
      end loop;
      return Res;
   end;

   function "abs"(Ui : in Unlimited_Static) return Unlimited_Static is
   begin
      if Ui(Ui'Last)<0 then
         return -Ui;
      else
         return Ui;
      end if;
   end;

   function PrepareResult(U1,U2 : in Unlimited_Static) return Unlimited_Static is
   begin
      return Extend_Range(U1,
         Integer'Min(U1'First,U2'First),
         Integer'Max(U1'Last,U2'Last)+1);
   end;

   function "+"(Left,Right : in Unlimited_Static) return Unlimited_Static is
      Res : Unlimited_Static := Prepareresult(Left,Right);
   begin
      Add(Res,Right);
      return Reduce(Res);
   end;

   function "-"(Left,Right : in Unlimited_Static) return Unlimited_Static is
      Res : Unlimited_Static := Prepareresult(Left,Right);
   begin
      Sub(Res,Right);
      return Reduce(Res);
   end;

   function "-"(Ui : in Unlimited_Static) return Unlimited_Static is
      Res : Unlimited_Static := not Ui;
   begin
      return Extend_Range(Res+(Ui'First=>1),Ui'First,Ui'Last);
   end;

   generic
      with function Bit_Operator(Ui1,Ui2 : UInt) return UInt;
   function Bit_Operation(U1,U2 : in Unlimited_Static) return Unlimited_Static;

   function Bit_Operation(U1,U2 : in Unlimited_Static) return Unlimited_Static is
      Result : Unlimited_Static := Extend_Range(U1,
         Integer'Min(U1'First,U2'First),Integer'Max(U1'Last,U2'Last));
      Sign2 : UInt := U(Sign_Part(U2));
   begin
      for I in Result'First..U2'First-1 loop
         Result(I) := P(Bit_Operator(U(Result(I)),0));
      end loop;
      for I in U2'range loop
         Result(I) := P(Bit_Operator(U(Result(I)),U(U2(I))));
      end loop;
      for I in U2'Last+1..Result'Last loop
         Result(I) := P(Bit_Operator(U(Result(I)),Sign2));
      end loop;
      return Reduce(Result);
   end;

   function And_Oper is new Bit_Operation("and");
   function "and"(Left,Right : in Unlimited_Static) return Unlimited_Static renames And_Oper;

   function Or_Oper is new Bit_Operation("or");
   function "or"(Left,Right : in Unlimited_Static) return Unlimited_Static renames Or_Oper;

   function Xor_Oper is new Bit_Operation("xor");
   function "xor"(Left,Right : in Unlimited_Static) return Unlimited_Static renames Xor_Oper;

end Unlimited.Number;
